Search.setIndex({docnames:["index","module_docs/haggle","module_docs/haggle/dacc","module_docs/haggle/small_test","table_of_contents"],envversion:{"sphinx.domains.c":2,"sphinx.domains.changeset":1,"sphinx.domains.citation":1,"sphinx.domains.cpp":3,"sphinx.domains.index":1,"sphinx.domains.javascript":2,"sphinx.domains.math":2,"sphinx.domains.python":2,"sphinx.domains.rst":2,"sphinx.domains.std":1,"sphinx.ext.todo":2,"sphinx.ext.viewcode":1,sphinx:56},filenames:["index.rst","module_docs/haggle.rst","module_docs/haggle/dacc.rst","module_docs/haggle/small_test.rst","table_of_contents.rst"],objects:{},objnames:{},objtypes:{},terms:{"2021":0,chang:0,content:[0,4],dacc:[0,4],dec:0,haggl:4,index:0,last:0,modul:0,page:0,releas:0,search:0,small_test:[0,4]},titles:["Welcome to haggle\u2019s documentation!","haggle","haggle.dacc","haggle.small_test","&lt;no title&gt;"],titleterms:{dacc:2,document:0,haggl:[0,1,2,3],indic:0,small_test:3,tabl:0,welcom:0}})